@@ -4,15 +4,18 @@ import logging
 import os
 import pathlib
 import subprocess
+from collections.abc import Iterator
 from contextlib import contextmanager
+from pathlib import Path
 from shutil import rmtree
-from typing import Any, Iterator, List, Type, Union
+from typing import Any, Union
 
 import pytest
 from _pytest.monkeypatch import MonkeyPatch
 from packaging.version import Version
 from pytest_mock import MockerFixture
 
+from ansible_compat.config import ansible_version
 from ansible_compat.constants import INVALID_PREREQUISITES_RC
 from ansible_compat.errors import (
     AnsibleCommandError,
@@ -47,8 +50,12 @@ def test_runtime_missing_ansible_module(monkeypatch: MonkeyPatch) -> None:
     class RaiseException:
         """Class to raise an exception."""
 
-        def __init__(self, *args: Any, **kwargs: Any) -> None:
-            raise ModuleNotFoundError()
+        def __init__(
+            self,
+            *args: Any,  # noqa: ARG002,ANN401
+            **kwargs: Any,  # noqa: ARG002,ANN401
+        ) -> None:
+            raise ModuleNotFoundError
 
     monkeypatch.setattr("importlib.import_module", RaiseException)
 
@@ -86,35 +93,40 @@ def test_runtime_version_fail_module(mocker: MockerFixture) -> None:
         autospec=True,
     )
     patched.side_effect = InvalidPrerequisiteError(
-        "Unable to parse ansible cli version"
+        "Unable to parse ansible cli version",
     )
     runtime = Runtime()
     with pytest.raises(
-        InvalidPrerequisiteError, match="Unable to parse ansible cli version"
+        InvalidPrerequisiteError,
+        match="Unable to parse ansible cli version",
     ):
-        runtime.version  # pylint: disable=pointless-statement
+        _ = runtime.version  # pylint: disable=pointless-statement
 
 
 def test_runtime_version_fail_cli(mocker: MockerFixture) -> None:
     """Tests for failure to detect Ansible version."""
     mocker.patch(
-        "ansible_compat.runtime.Runtime.exec",
+        "ansible_compat.runtime.Runtime.run",
         return_value=CompletedProcess(
-            ["x"], returncode=123, stdout="oops", stderr="some error"
+            ["x"],
+            returncode=123,
+            stdout="oops",
+            stderr="some error",
         ),
         autospec=True,
     )
     runtime = Runtime()
     with pytest.raises(
-        RuntimeError, match="Unable to find a working copy of ansible executable."
+        RuntimeError,
+        match="Unable to find a working copy of ansible executable.",
     ):
-        runtime.version  # pylint: disable=pointless-statement
+        _ = runtime.version  # pylint: disable=pointless-statement
 
 
 def test_runtime_prepare_ansible_paths_validation() -> None:
     """Check that we validate collection_path."""
     runtime = Runtime()
-    runtime.config.collections_paths = "invalid-value"  # type: ignore
+    runtime.config.collections_paths = "invalid-value"  # type: ignore[assignment]
     with pytest.raises(RuntimeError, match="Unexpected ansible configuration"):
         runtime._prepare_ansible_paths()
 
@@ -137,18 +149,18 @@ def test_runtime_install_role(
 ) -> None:
     """Checks that we can install roles."""
     caplog.set_level(logging.INFO)
-    project_dir = os.path.join(os.path.dirname(__file__), "roles", folder)
+    project_dir = Path(__file__).parent / "roles" / folder
     runtime = Runtime(isolated=isolated, project_dir=project_dir)
     runtime.prepare_environment(install_local=True)
     # check that role appears as installed now
-    result = runtime.exec(["ansible-galaxy", "list"])
+    result = runtime.run(["ansible-galaxy", "list"])
     assert result.returncode == 0, result
     assert role_name in result.stdout
     if isolated:
         assert pathlib.Path(f"{runtime.cache_dir}/roles/{role_name}").is_symlink()
     else:
         assert pathlib.Path(
-            f"{os.path.expanduser(runtime.config.default_roles_path[0])}/{role_name}"
+            f"{Path(runtime.config.default_roles_path[0]).expanduser()}/{role_name}",
         ).is_symlink()
     runtime.clean()
     # also test that clean does not break when cache_dir is missing
@@ -160,7 +172,7 @@ def test_runtime_install_role(
 
 def test_prepare_environment_with_collections(tmp_path: pathlib.Path) -> None:
     """Check that collections are correctly installed."""
-    runtime = Runtime(isolated=True, project_dir=str(tmp_path))
+    runtime = Runtime(isolated=True, project_dir=tmp_path)
     runtime.prepare_environment(required_collections={"community.molecule": "0.1.0"})
 
 
@@ -168,32 +180,24 @@ def test_runtime_install_requirements_missing_file() -> None:
     """Check that missing requirements file is ignored."""
     # Do not rely on this behavior, it may be removed in the future
     runtime = Runtime()
-    runtime.install_requirements("/that/does/not/exist")
+    runtime.install_requirements(Path("/that/does/not/exist"))
 
 
 @pytest.mark.parametrize(
     ("file", "exc", "msg"),
     (
         (
-            "/dev/null",
+            Path("/dev/null"),
             InvalidPrerequisiteError,
             "file is not a valid Ansible requirements file",
         ),
         (
-            os.path.join(
-                os.path.dirname(__file__),
-                "assets",
-                "requirements-invalid-collection.yml",
-            ),
+            Path(__file__).parent / "assets" / "requirements-invalid-collection.yml",
             AnsibleCommandError,
             "Got 1 exit code while running: ansible-galaxy",
         ),
         (
-            os.path.join(
-                os.path.dirname(__file__),
-                "assets",
-                "requirements-invalid-role.yml",
-            ),
+            Path(__file__).parent / "assets" / "requirements-invalid-role.yml",
             AnsibleCommandError,
             "Got 1 exit code while running: ansible-galaxy",
         ),
@@ -201,7 +205,9 @@ def test_runtime_install_requirements_missing_file() -> None:
     ids=("empty", "invalid-collection", "invalid-role"),
 )
 def test_runtime_install_requirements_invalid_file(
-    file: str, exc: Type[Any], msg: str
+    file: Path,
+    exc: type[Any],
+    msg: str,
 ) -> None:
     """Check that invalid requirements file is raising."""
     runtime = Runtime()
@@ -213,26 +219,21 @@ def test_runtime_install_requirements_invalid_file(
 
 
 @contextmanager
-def remember_cwd(cwd: str) -> Iterator[None]:
-    """Context manager for chdir."""
-    curdir = os.getcwd()
+def cwd(path: Path) -> Iterator[None]:
+    """Context manager for temporary changing current working directory."""
+    old_pwd = Path.cwd()
+    os.chdir(path)
     try:
-        os.chdir(cwd)
         yield
     finally:
-        os.chdir(curdir)
+        os.chdir(old_pwd)
 
 
 def test_prerun_reqs_v1(caplog: pytest.LogCaptureFixture, runtime: Runtime) -> None:
     """Checks that the linter can auto-install requirements v1 when found."""
-    cwd = os.path.realpath(
-        os.path.join(
-            os.path.dirname(os.path.realpath(__file__)), "..", "examples", "reqs_v1"
-        )
-    )
-    with remember_cwd(cwd):
-        with caplog.at_level(logging.INFO):
-            runtime.prepare_environment()
+    path = Path(__file__).parent.parent / "examples" / "reqs_v1"
+    with cwd(path), caplog.at_level(logging.INFO):
+        runtime.prepare_environment()
     assert any(
         msg.startswith("Running ansible-galaxy role install") for msg in caplog.messages
     )
@@ -244,12 +245,8 @@ def test_prerun_reqs_v1(caplog: pytest.LogCaptureFixture, runtime: Runtime) -> N
 
 def test_prerun_reqs_v2(caplog: pytest.LogCaptureFixture, runtime: Runtime) -> None:
     """Checks that the linter can auto-install requirements v2 when found."""
-    cwd = os.path.realpath(
-        os.path.join(
-            os.path.dirname(os.path.realpath(__file__)), "..", "examples", "reqs_v2"
-        )
-    )
-    with remember_cwd(cwd):
+    path = (Path(__file__).parent.parent / "examples" / "reqs_v2").resolve()
+    with cwd(path):
         with caplog.at_level(logging.INFO):
             runtime.prepare_environment()
         assert any(
@@ -262,6 +259,13 @@ def test_prerun_reqs_v2(caplog: pytest.LogCaptureFixture, runtime: Runtime) -> N
         )
 
 
+def test_prerun_reqs_broken(runtime: Runtime) -> None:
+    """Checks that the we report invalid requirements.yml file."""
+    path = (Path(__file__).parent.parent / "examples" / "reqs_broken").resolve()
+    with cwd(path), pytest.raises(InvalidPrerequisiteError):
+        runtime.prepare_environment()
+
+
 def test__update_env_no_old_value_no_default_no_value(monkeypatch: MonkeyPatch) -> None:
     """Make sure empty value does not touch environment."""
     monkeypatch.delenv("DUMMY_VAR", raising=False)
@@ -301,7 +305,9 @@ def test__update_env_no_default_no_value(monkeypatch: MonkeyPatch) -> None:
     ),
 )
 def test__update_env_no_old_value_no_default(
-    monkeypatch: MonkeyPatch, value: List[str], result: str
+    monkeypatch: MonkeyPatch,
+    value: list[str],
+    result: str,
 ) -> None:
     """Values are concatenated using : as the separator."""
     monkeypatch.delenv("DUMMY_VAR", raising=False)
@@ -320,7 +326,10 @@ def test__update_env_no_old_value_no_default(
     ),
 )
 def test__update_env_no_old_value(
-    monkeypatch: MonkeyPatch, default: str, value: List[str], result: str
+    monkeypatch: MonkeyPatch,
+    default: str,
+    value: list[str],
+    result: str,
 ) -> None:
     """Values are appended to default value."""
     monkeypatch.delenv("DUMMY_VAR", raising=False)
@@ -339,7 +348,10 @@ def test__update_env_no_old_value(
     ),
 )
 def test__update_env_no_default(
-    monkeypatch: MonkeyPatch, old_value: str, value: List[str], result: str
+    monkeypatch: MonkeyPatch,
+    old_value: str,
+    value: list[str],
+    result: str,
 ) -> None:
     """Values are appended to preexisting value."""
     monkeypatch.setenv("DUMMY_VAR", old_value)
@@ -362,8 +374,8 @@ def test__update_env_no_default(
 def test__update_env(
     monkeypatch: MonkeyPatch,
     old_value: str,
-    default: str,  # pylint: disable=unused-argument
-    value: List[str],
+    default: str,  # pylint: disable=unused-argument # noqa: ARG001
+    value: list[str],
     result: str,
 ) -> None:
     """Defaults are ignored when preexisting value is present."""
@@ -378,14 +390,14 @@ def test__update_env(
 def test_require_collection_wrong_version(runtime: Runtime) -> None:
     """Tests behaviour of require_collection."""
     subprocess.check_output(
-        [
+        [  # noqa: S603
             "ansible-galaxy",
             "collection",
             "install",
             "examples/reqs_v2/community-molecule-0.1.0.tar.gz",
             "-p",
             "~/.ansible/collections",
-        ]
+        ],
     )
     with pytest.raises(InvalidPrerequisiteError) as pytest_wrapped_e:
         runtime.require_collection("community.molecule", "9999.9.9")
@@ -396,26 +408,28 @@ def test_require_collection_wrong_version(runtime: Runtime) -> None:
 def test_require_collection_invalid_name(runtime: Runtime) -> None:
     """Check that require_collection raise with invalid collection name."""
     with pytest.raises(
-        InvalidPrerequisiteError, match="Invalid collection name supplied:"
+        InvalidPrerequisiteError,
+        match="Invalid collection name supplied:",
     ):
         runtime.require_collection("that-is-invalid")
 
 
 def test_require_collection_invalid_collections_path(runtime: Runtime) -> None:
     """Check that require_collection raise with invalid collections path."""
-    runtime.config.collections_paths = "/that/is/invalid"  # type: ignore
+    runtime.config.collections_paths = "/that/is/invalid"  # type: ignore[assignment]
     with pytest.raises(
-        InvalidPrerequisiteError, match="Unable to determine ansible collection paths"
+        InvalidPrerequisiteError,
+        match="Unable to determine ansible collection paths",
     ):
         runtime.require_collection("community.molecule")
 
 
 def test_require_collection_preexisting_broken(tmp_path: pathlib.Path) -> None:
     """Check that require_collection raise with broken pre-existing collection."""
-    runtime = Runtime(isolated=True, project_dir=str(tmp_path))
+    runtime = Runtime(isolated=True, project_dir=tmp_path)
     dest_path: str = runtime.config.collections_paths[0]
-    dest = os.path.join(dest_path, "ansible_collections", "foo", "bar")
-    os.makedirs(dest, exist_ok=True)
+    dest = pathlib.Path(dest_path) / "ansible_collections" / "foo" / "bar"
+    dest.mkdir(parents=True, exist_ok=True)
     with pytest.raises(InvalidPrerequisiteError, match="missing MANIFEST.json"):
         runtime.require_collection("foo.bar")
 
@@ -435,7 +449,10 @@ def test_require_collection(runtime_tmp: Runtime) -> None:
     ids=("a", "b", "c"),
 )
 def test_require_collection_missing(
-    name: str, version: str, install: bool, runtime: Runtime
+    name: str,
+    version: str,
+    install: bool,
+    runtime: Runtime,
 ) -> None:
     """Tests behaviour of require_collection, missing case."""
     with pytest.raises(AnsibleCompatError) as pytest_wrapped_e:
@@ -452,7 +469,8 @@ def test_install_collection(runtime: Runtime) -> None:
 def test_install_collection_dest(runtime: Runtime, tmp_path: pathlib.Path) -> None:
     """Check that valid collection to custom destination passes."""
     runtime.install_collection(
-        "examples/reqs_v2/community-molecule-0.1.0.tar.gz", destination=tmp_path
+        "examples/reqs_v2/community-molecule-0.1.0.tar.gz",
+        destination=tmp_path,
     )
     expected_file = (
         tmp_path / "ansible_collections" / "community" / "molecule" / "MANIFEST.json"
@@ -479,13 +497,15 @@ def test_install_galaxy_role(runtime_tmp: Runtime) -> None:
     runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=2)
     # this should raise an error
     with pytest.raises(
-        InvalidPrerequisiteError, match="does not follow current galaxy requirements"
+        InvalidPrerequisiteError,
+        match="does not follow current galaxy requirements",
     ):
         runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=0)
 
 
 def test_install_galaxy_role_unlink(
-    runtime_tmp: Runtime, caplog: pytest.LogCaptureFixture
+    runtime_tmp: Runtime,
+    caplog: pytest.LogCaptureFixture,
 ) -> None:
     """Test ability to unlink incorrect symlinked roles."""
     caplog.set_level(logging.INFO)
@@ -506,14 +526,13 @@ def test_install_galaxy_role_unlink(
 
 def test_install_galaxy_role_bad_namespace(runtime_tmp: Runtime) -> None:
     """Check install role with bad namespace in galaxy info."""
-    # pathlib.Path(f'{runtime_tmp.project_dir}/galaxy.yml').touch()
     pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir()
     pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").write_text(
         """galaxy_info:
   role_name: foo
   author: bar
   namespace: ["xxx"]
-"""
+""",
     )
     # this should raise an error regardless the role_name_check value
     with pytest.raises(AnsibleCompatError, match="Role namespace must be string, not"):
@@ -559,10 +578,10 @@ def test_install_galaxy_role_no_checks(runtime_tmp: Runtime) -> None:
   role_name: foo
   author: bar
   namespace: acme
-"""
+""",
     )
     runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=2)
-    result = runtime_tmp.exec(["ansible-galaxy", "list"])
+    result = runtime_tmp.run(["ansible-galaxy", "list"])
     assert "- acme.foo," in result.stdout
     assert result.returncode == 0, result
 
@@ -614,7 +633,9 @@ def test_runtime_env_ansible_library(monkeypatch: MonkeyPatch) -> None:
     ids=("1", "2", "3", "4", "5"),
 )
 def test_runtime_version_in_range(
-    lower: Union[str, None], upper: Union[str, None], expected: bool
+    lower: Union[str, None],
+    upper: Union[str, None],
+    expected: bool,
 ) -> None:
     """Validate functioning of version_in_range."""
     runtime = Runtime()
@@ -622,35 +643,59 @@ def test_runtime_version_in_range(
 
 
 @pytest.mark.parametrize(
-    ("path", "scenario"),
+    ("path", "scenario", "expected_collections"),
     (
-        ("test/collections/acme.goodies", "default"),
-        ("test/collections/acme.goodies/roles/baz", "deep_scenario"),
+        pytest.param(
+            "test/collections/acme.goodies",
+            "default",
+            [
+                "ansible.posix",  # from tests/requirements.yml
+                "ansible.utils",  # from galaxy.yml
+                "community.molecule",  # from galaxy.yml
+            ],
+            id="normal",
+        ),
+        pytest.param(
+            "test/collections/acme.goodies/roles/baz",
+            "deep_scenario",
+            ["community.molecule"],
+            id="deep",
+        ),
     ),
-    ids=("normal", "deep"),
 )
-def test_install_collection_from_disk(path: str, scenario: str) -> None:
+def test_install_collection_from_disk(
+    path: str,
+    scenario: str,
+    expected_collections: list[str],
+) -> None:
     """Tests ability to install a local collection."""
-    # ensure we do not have acme.google installed in user directory as it may
+    # ensure we do not have acme.goodies installed in user directory as it may
     # produce false positives
     rmtree(
-        os.path.expanduser("~/.ansible/collections/ansible_collections/acme/goodies"),
+        pathlib.Path(
+            "~/.ansible/collections/ansible_collections/acme/goodies",
+        ).expanduser(),
         ignore_errors=True,
     )
-    with remember_cwd(path):
+    with cwd(Path(path)):
         runtime = Runtime(isolated=True)
         # this should call install_collection_from_disk(".")
         runtime.prepare_environment(install_local=True)
         # that molecule converge playbook can be used without molecule and
         # should validate that the installed collection is available.
-        result = runtime.exec(["ansible-playbook", f"molecule/{scenario}/converge.yml"])
+        result = runtime.run(["ansible-playbook", f"molecule/{scenario}/converge.yml"])
         assert result.returncode == 0, result.stdout
+        runtime.load_collections()
+        for collection_name in expected_collections:
+            assert (
+                collection_name in runtime.collections
+            ), f"{collection_name} not found in {runtime.collections.keys()}"
         runtime.clean()
 
 
 def test_install_collection_from_disk_fail() -> None:
     """Tests that we fail to install a broken collection."""
-    with remember_cwd("test/collections/acme.broken"):
+    with cwd(Path("test/collections/acme.broken")):
         runtime = Runtime(isolated=True)
         with pytest.raises(RuntimeError) as exc_info:
             runtime.prepare_environment(install_local=True)
@@ -663,21 +708,21 @@ def test_install_collection_from_disk_fail() -> None:
             InvalidPrerequisiteError,
         )
         assert exc_info.match(
-            "(is missing the following mandatory|Got 1 exit code while running: ansible-galaxy collection build)"
+            "(is missing the following mandatory|Got 1 exit code while running: ansible-galaxy collection build)",
         )
 
 
 def test_prepare_environment_offline_role() -> None:
     """Ensure that we can make use of offline roles."""
-    with remember_cwd("test/roles/acme.missing_deps"):
+    with cwd(Path("test/roles/acme.missing_deps")):
         runtime = Runtime(isolated=True)
         runtime.prepare_environment(install_local=True, offline=True)
 
 
 def test_runtime_run(runtime: Runtime) -> None:
     """Check if tee and non tee mode return same kind of results."""
-    result1 = runtime.exec(["seq", "10"])
-    result2 = runtime.exec(["seq", "10"], tee=True)
+    result1 = runtime.run(["seq", "10"])
+    result2 = runtime.run(["seq", "10"], tee=True)
     assert result1.returncode == result2.returncode
     assert result1.stderr == result2.stderr
     assert result1.stdout == result2.stdout
@@ -685,21 +730,62 @@ def test_runtime_run(runtime: Runtime) -> None:
 
 def test_runtime_exec_cwd(runtime: Runtime) -> None:
     """Check if passing cwd works as expected."""
-    cwd = "/"
-    result1 = runtime.exec(["pwd"], cwd=cwd)
-    result2 = runtime.exec(["pwd"])
-    assert result1.stdout.rstrip() == cwd
+    path = Path("/")
+    result1 = runtime.run(["pwd"], cwd=path)
+    result2 = runtime.run(["pwd"])
+    assert result1.stdout.rstrip() == str(path)
     assert result1.stdout != result2.stdout
 
 
 def test_runtime_exec_env(runtime: Runtime) -> None:
     """Check if passing env works."""
-    result = runtime.exec(["printenv", "FOO"])
+    result = runtime.run(["printenv", "FOO"])
     assert not result.stdout
 
-    result = runtime.exec(["printenv", "FOO"], env={"FOO": "bar"})
+    result = runtime.run(["printenv", "FOO"], env={"FOO": "bar"})
     assert result.stdout.rstrip() == "bar"
 
     runtime.environ["FOO"] = "bar"
-    result = runtime.exec(["printenv", "FOO"])
+    result = runtime.run(["printenv", "FOO"])
     assert result.stdout.rstrip() == "bar"
+
+
+def test_runtime_plugins(runtime: Runtime) -> None:
+    """Tests ability to access detected plugins."""
+    assert len(runtime.plugins.cliconf) == 0
+    # ansible.netcommon.restconf might be in httpapi
+    assert isinstance(runtime.plugins.httpapi, dict)
+    # "ansible.netcommon.default" might be in runtime.plugins.netconf
+    assert isinstance(runtime.plugins.netconf, dict)
+    assert isinstance(runtime.plugins.role, dict)
+    assert "become" in runtime.plugins.keyword
+
+    if ansible_version() < Version("2.14.0"):
+        assert "sudo" in runtime.plugins.become
+        assert "memory" in runtime.plugins.cache
+        assert "default" in runtime.plugins.callback
+        assert "local" in runtime.plugins.connection
+        assert "ini" in runtime.plugins.inventory
+        assert "env" in runtime.plugins.lookup
+        assert "sh" in runtime.plugins.shell
+        assert "host_group_vars" in runtime.plugins.vars
+        assert "file" in runtime.plugins.module
+        assert "free" in runtime.plugins.strategy
+        # ansible-doc below 2.14 does not support listing 'test' and 'filter' types:
+        with pytest.raises(RuntimeError):
+            assert "is_abs" in runtime.plugins.test
+        with pytest.raises(RuntimeError):
+            assert "bool" in runtime.plugins.filter
+    else:
+        assert "ansible.builtin.sudo" in runtime.plugins.become
+        assert "ansible.builtin.memory" in runtime.plugins.cache
+        assert "ansible.builtin.default" in runtime.plugins.callback
+        assert "ansible.builtin.local" in runtime.plugins.connection
+        assert "ansible.builtin.ini" in runtime.plugins.inventory
+        assert "ansible.builtin.env" in runtime.plugins.lookup
+        assert "ansible.builtin.sh" in runtime.plugins.shell
+        assert "ansible.builtin.host_group_vars" in runtime.plugins.vars
+        assert "ansible.builtin.file" in runtime.plugins.module
+        assert "ansible.builtin.free" in runtime.plugins.strategy
+        assert "ansible.builtin.is_abs" in runtime.plugins.test
+        assert "ansible.builtin.bool" in runtime.plugins.filter
